"""
These are the items that can be added to the staff toolbar.
"""
import sys

from django.contrib.admin.templatetags.admin_urls import admin_urlname
from django.core.urlresolvers import reverse, reverse_lazy
from django.utils.encoding import force_text
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _

if sys.version_info[0] >= 3:
    basestring = str


__all__ = (
    'Group',
    'RootNode',
    'Link',
    'AdminLink',
    'ObjectNode',
)


class Title(object):
    """
    A title in the toolbar.
    """
    def __init__(self, title):
        self.title = title

    def __call__(self, request, context):
        return format_html(u'<div class="toolbar-title">{0}</div>', self.title)


class Literal(object):
    """
    A literal object, that is outputted as-is.
    Use ``format_html()`` or ``mark_safe()`` to output literal HTML.
    """
    def __init__(self, title):
        self.title = title

    def __call__(self, request, context):
        return self.title


class Group(object):
    """
    A group of items
    """
    title = None

    def __init__(self, *children, **kwargs):
        self.children = list(children)
        self.title = kwargs.get('title', self.title)

    def __call__(self, request, context):
        """
        Render the group
        """
        rows = self.get_rows(request, context)
        return self.render(rows)

    def get_rows(self, request, context):
        """
        Get all rows as HTML
        """
        from staff_toolbar.loading import load_toolbar_item
        rows = []
        for i, hook in enumerate(self.children):
            # Allow dotted paths in groups too, loads on demand (get import errors otherwise).
            if isinstance(hook, (basestring, tuple, list)):
                hook = load_toolbar_item(hook)
                self.children[i] = hook

            html = hook(request, context)
            if not html:
                continue

            rows.append(html)
        return rows

    def render(self, rows):
        """
        Join the HTML rows.
        """
        if not rows:
            return ''

        li_tags = mark_safe(u"\n".join(format_html(u'<li>{0}</li>', force_text(row)) for row in rows))
        if self.title:
            return format_html(u'<div class="toolbar-title">{0}</div>\n<ul>\n{1}\n</ul>', self.title, li_tags)
        else:
            return format_html(u'<ul>\n{0}\n</ul>', li_tags)


class RootNode(Group):
    """
    The root node for the toolbar.
    """
    title = _("Staff features")



class Link(object):
    """
    Add a hard-coded link in the toolbar.
    """
    # Allow to define class-level defaults.
    url = None
    title = None

    def __init__(self, url=None, title=None):
        # Assign class-level values to 'self' if nothing is passed.
        self.url = url or self.url
        self.title = title or self.title

    def __call__(self, request, context):
        linkdata = self.get_link(request, context)
        if linkdata:
            return format_html(u'<a href="{0}">{1}</a>', *linkdata)
        else:
            return None

    def get_link(self, request, context):
        return (self.url, self.title)


class AdminIndexLink(Link):
    """
    A link to the admin index.
    """
    url = reverse_lazy('admin:index')
    title = _("Admin dashboard")


class ChangeObjectLink(Link):
    """
    Display a link to the admin of the current page.
    It tries to fetch the URL using:

    * ``{% set_staff_object ... %}``
    * ``{% set_staff_url %}...{% end_set_staff_url %}``
    * ``request.staff_url``
    * ``request.staff_object``
    * ``view.get_staff_object()`
    * ``view.get_staff_url()`
    * ``view.object``
    * ``context['object']``
    """
    def get_link(self, request, context):
        # When `set_staff_object` is used, take that information,
        object = getattr(request, 'staff_object', None)
        url = getattr(request, 'staff_url', None)

        if not object and not url:
            # If the information was not passed in the template,
            # the view can also be queried.
            view = context.get('view')
            if view is not None:
                # Implementing StaffUrlMixin
                if hasattr(view, 'get_staff_url'):
                    url = view.get_staff_url()
                if hasattr(view, 'get_staff_object'):
                    object = view.get_staff_object()

            if not object:
                # Last resort, try to find the object in the context or view.
                object = get_object(context)

        if object and url:
            # URL and object, use that!
            return (url, _admin_title(object))
        if object:
            # No URL, take default
            return (_admin_url(object), _admin_title(object))
        if url:
            # URL only, default title
            return (url, _("Edit content"))

        return None


class LogoutLink(Link):
    """
    A logoff link.
    """
    url = reverse_lazy('admin:logout')
    title = _("Logout")



def get_object(context):
    """
    Get an object from the context or view.
    """
    object = None

    view = context.get('view')
    if view:
        # View is more reliable then an 'object' variable in the context.
        # Works if this is a SingleObjectMixin
        object = getattr(view, 'object', None)

    if object is None:
        object = context.get('object', None)

    return object


def _admin_url(object):
    url = reverse(admin_urlname(object._meta, 'change'), args=(object.pk,))
    if hasattr(object, '_parler_meta'):
        # django-parler's TranslatableModel object.
        # Open the admin with the current language tab.
        return u"{0}?language={1}".format(url, object.get_current_language())
    else:
        return url


def _admin_title(object):
    return _('Change %s') % force_text(object._meta.verbose_name)



from django.contrib.auth.models import AnonymousUser
from django.test import RequestFactory
from django.test import TestCase
from staff_toolbar import appsettings
from staff_toolbar.loading import get_toolbar_root


def get_dummy_request():
    """
    Returns a Request instance.
    """
    request = RequestFactory().get("/", HTTP_HOST='example.org')
    request.session = {}
    request.user = AnonymousUser()
    return request


class StaffToolbarTests(TestCase):
    """
    Staff toolbar test
    """

    def test_loading(self):
        root = get_toolbar_root()

        self.assertEqual(len(root.children), len(appsettings.STAFF_TOOLBAR_ITEMS))

        request = get_dummy_request()
        html = root(request, {})
        self.assertHTMLEqual(
            html,
            '''
            <div class="toolbar-title">Staff features</div>
            <ul>
                <li><a href="/admin/">Admin dashboard</a></li>
                <li><a href="/admin/logout/">Logout</a></li>
            </ul>
            '''
        )



from django.template import Library, Node
from django.utils.html import format_html

from staff_toolbar.loading import get_toolbar_root

register = Library()

toolbar_root = get_toolbar_root()


@register.simple_tag(takes_context=True)
def staff_toolbar(context):
    """
    Display the staff toolbar
    :param context:
    :type context:
    :return:
    :rtype:
    """
    request = context['request']
    if not request.user.is_staff:
        return u''

    toolbar_html = toolbar_root(request, context)
    return format_html(u'<nav id="django-staff-toolbar">{0}</nav>', toolbar_html)


@register.simple_tag(takes_context=True)
def set_staff_object(context, object):
    """
    Assign an object to be the "main object" of this page.
    Example::

        {% set_staff_object page %}
    """
    request = context['request']
    request.staff_object = object
    return u''


@register.tag
def set_staff_url(parser, token):
    """
    Assign an URL to be the "admin link" of this page.
    Example::

        {% set_staff_url %}{% url 'admin:fluent_pages_page_change' page.id %}{% end_set_staff_url %}
    """
    nodelist = parser.parse(('end_set_staff_url',))
    parser.delete_first_token()
    return AdminUrlNode(nodelist)


class AdminUrlNode(Node):
    def __init__(self, nodelist):
        self.nodelist = nodelist

    def render(self, context):
        url = self.nodelist.render(context)
        context['request'].staff_url = url
        return u''





def toolbar_title(title):
    """
    Define a title to be included in the toolbar.
    """
    return LazyToolbarItem('staff_toolbar.items.Title', title)


def toolbar_literal(title):
    """
    Define a literal text to be included in the toolbar.
    """
    return LazyToolbarItem('staff_toolbar.items.Literal', title)


def toolbar_item(callable_path, *args, **kwargs):
    """
    Defining a toolbar item in the settings that also received arguments.
    It won't be loaded until the toolbar is actually rendered.
    """
    return LazyToolbarItem(callable_path, *args, **kwargs)


def toolbar_link(url, title):
    """
    Define a link to be included in the toolbar.
    """
    return LazyToolbarItem('staff_toolbar.items.Link', url=url, title=title)


class LazyToolbarItem(object):
    """
    Internal mechanism to support lazy-loaded toolbar items in the settings.
    """
    def __init__(self, import_path, *args, **kwargs):
        self.import_path = import_path
        self.args = args
        self.kwargs = kwargs
        self.real_instance = None

    def __call__(self, request, context):
        if self.real_instance is None:
            # Init on demand.
            from staff_toolbar.loading import load_toolbar_item
            self.real_instance = load_toolbar_item(self.import_path, *self.args, **self.kwargs)

        return self.real_instance(request, context)



from django.contrib.admin.templatetags.admin_urls import admin_urlname
from django.core.urlresolvers import reverse


class StaffUrlMixin(object):
    """
    Provide a ``view.get_admin_url`` variable in the template.
    """
    def get_staff_object(self):
        return getattr(self, 'object', None)

    def get_staff_url(self):
        """
        Return the Admin URL for the current view.
        By default, it uses the :func:`get_staff_object` function to base the URL on.
        """
        object = self.get_staff_object()
        if object is not None:
            # View is likely using SingleObjectMixin
            return reverse(admin_urlname(object._meta, 'change'), args=(object.pk,))

        model = _get_view_model(self)
        if model is not None:
            # View is likely using MultipleObjectMixin (e.g. ListView)
            return reverse(admin_urlname(object._meta, 'changelist'))

        return None



def _get_view_model(view):
    if view.model is not None:
        # If a model has been explicitly provided, use it
        return view.model
    elif hasattr(view, 'object') and view.object is not None:
        # If this view is operating on a single object, use the class of that object
        return view.object.__class__
    elif hasattr(view, 'get_queryset'):
        # Try to get a queryset and extract the model class from that
        return view.get_queryset().model
    else:
        return None



from django.conf import settings

STAFF_TOOLBAR_ITEMS = getattr(settings, 'STAFF_TOOLBAR_ITEMS', (
    'staff_toolbar.items.AdminIndexLink',
    'staff_toolbar.items.ChangeObjectLink',
    'staff_toolbar.items.LogoutLink',
))



"""
The internal machinery to load the toolbar items.
"""
import inspect
import traceback
import sys
from django.core.exceptions import ImproperlyConfigured
from staff_toolbar import appsettings
from staff_toolbar.items import RootNode, Group

try:
    from importlib import import_module
except ImportError:
    from django.utils.importlib import import_module  # Python 2.6 compatibility

if sys.version_info[0] >= 3:
    basestring = str

_toolbar_root = None


__all__ = (
    'get_toolbar_root',
    'load_toolbar_item'
)


def get_toolbar_root():
    """
    Init on demand.
    :rtype: RootNode
    """
    global _toolbar_root
    if _toolbar_root is None:
        items = [load_toolbar_item(item) for item in appsettings.STAFF_TOOLBAR_ITEMS]
        _toolbar_root = RootNode(*items)
    return _toolbar_root


def load_toolbar_item(import_path, *args, **kwargs):
    """
    Load an item in the toolbar
    :param import_path: the dotted python path to class or function.
    :param args: For classes, any arguments to pass to the constructor.
    :param kwargs: For classes, any keyword arguments to pass to the constructor.
    """
    if isinstance(import_path, (tuple, list)):
        children = [load_toolbar_item(path) for path in import_path]
        return Group(*children)
    elif isinstance(import_path, basestring):
        symbol = _import_symbol(import_path, 'STAFF_TOOLBAR_ITEMS')
    else:
        symbol = import_path

    if inspect.isclass(symbol):
        # Instantiate the class.
        symbol = symbol(*args, **kwargs)

    if not callable(symbol):
        raise ImproperlyConfigured("The {0} in {1} is not callable!".format(import_path, 'STAFF_TOOLBAR_ITEMS'))

    return symbol


def _import_symbol(import_path, setting_name):
    """
    Import a class or function by name.
    """
    mod_name, class_name = import_path.rsplit('.', 1)

    # import module
    try:
        mod = import_module(mod_name)
        cls = getattr(mod, class_name)
    except ImportError as e:
        __, __, exc_traceback = sys.exc_info()
        frames = traceback.extract_tb(exc_traceback)
        if len(frames) > 1:
            raise   # import error is a level deeper.

        raise ImproperlyConfigured("{0} does not point to an existing class: {1}".format(setting_name, import_path))
    except AttributeError:
        raise ImproperlyConfigured("{0} does not point to an existing class: {1}".format(setting_name, import_path))

    return cls



from .conf import toolbar_title, toolbar_item, toolbar_literal, toolbar_link

__version__ = "1.1.1"  # following PEP 440

__all__ = (
    'toolbar_title',
    'toolbar_item',
    'toolbar_literal',
    'toolbar_link',
)
